import selection_strategy
import torchvision
from robustness import data_augmentation as da
import poisoned_datasets
from functools import partial
import poisoning_adversary
import torch
import numpy as np
import gammas
from typing import List, Dict, Callable, Union
import pickle
from argparse import ArgumentParser
import os

'''
Main function in this module: poisoning_function_from_args

This module consists of various combinations that might be useful for poisoning.
Eg. (hard_greedy, input_space_selector, max_degree_gamma_selection) All
functions take in points, and return a mask
'''

def projected_space_degree_bound_gamma(
    projected_dim: int,
    budget: int,
    gamma_selection_algorithm: Callable[[List[int]], Union[int, float]],
    point_selection_algorithm,
    degree_limit: int
    ):
    batch_size = 128
    def selection_algorithm(points):
        projected_points = sklearn.decomposition.PCA(projected_dim).fit_transform(points)
        gamma = gammas.search_degree_gamma(
            points=projected_points,
            degree_limit=degree_limit,
            lower_limit=0.0,
            upper_limit=1.0,
            iterations=10,
            norm=float('inf'),
            batch_size=batch_size,
            degree_func=gamma_selection_algorithm,
            use_tqdm=True
        )
        return poisoning_adversary.projected_space_selector(
            points=points,
            projected_dim=projected_dim,
            norm=float('inf'),
            norm_threshold=gamma[0],
            selection_algorithm=point_selection_algorithm,
            batch_size=batch_size,
            budgets=[budget]
        )[budget]
    return selection_algorithm


def input_space_degree_bound_gamma(
    budget: int,
    gamma_selection_algorithm: Callable[[List[int]], Union[int, float]],
    point_selection_algorithm,
    degree_limit: int,
    gamma_limits,
    iterations
    ):
    batch_size = 128
    def selection_algorithm(points):
        gamma = gammas.search_degree_gamma(
            points=points,
            degree_limit=degree_limit,
            lower_limit=gamma_limits[0],
            upper_limit=gamma_limits[1],
            iterations=iterations,
            norm=float('inf'),
            batch_size=batch_size,
            degree_func=gamma_selection_algorithm,
            use_tqdm=True
        )
        return poisoning_adversary.input_space_selector(
            points=points,
            norm=float('inf'),
            norm_threshold=gamma[0],
            selection_algorithm=point_selection_algorithm,
            batch_size=batch_size,
            budgets=[budget]
        )[budget]
    return selection_algorithm

def feature_space_degree_bound_gamma(
    feature_extractor,
    budget: int,
    gamma_selection_algorithm: Callable[[List[int]], Union[int, float]],
    point_selection_algorithm,
    degree_limit: int
    ):
    batch_size = 128
    def selection_algorithm(points):
        projected_points = poisoning_adversary._batched_feature_extractor(points, feature_extractor)
        gamma = gammas.search_degree_gamma(
            points=projected_points,
            degree_limit=degree_limit,
            lower_limit=0.0,
            upper_limit=1.0,
            iterations=10,
            norm=float('inf'),
            batch_size=batch_size,
            degree_func=gamma_selection_algorithm,
            use_tqdm=True
        )
        return poisoning_adversary.feature_space_selector(
            points=points,
            feature_extractor=feature_extractor,
            norm=float('inf'),
            norm_threshold=gamma[0],
            selection_algorithm=point_selection_algorithm,
            batch_size=batch_size,
            budgets=[budget]
        )[budget]
    return selection_algorithm



def random_space(
    budget: int
    ):
    return lambda points: poisoning_adversary.random_selector(
        points=points, budgets=[budget]
    )[budget]


def _get_poisoning_function_input_or_projected(
    budget: int,
    gamma_selection_algorithm: str,
    point_selection_algorithm: str,
    degree_limit: int,
    mode: str,
    projected_dim: int = None,
    degree_range=None
    ):
    projected = (mode == 'projected')
    assert(not projected or projected and projected_dim is not None)
    assert mode in ['projected', 'input']
    gamma_selection_algorithm = gammas.GAMMA_FUNCTIONS[gamma_selection_algorithm]
    if point_selection_algorithm != 'degree_range_random':
        point_selection_algorithm = selection_strategy.STRATEGIES[point_selection_algorithm]
    else:
        point_selection_algorithm = partial(
            selection_strategy.STRATEGIES[point_selection_algorithm], 
            degree_range
        )
    if mode == 'projected':
        result = partial(projected_space_degree_bound_gamma, projected_dim)
    elif mode == 'input':
        result = input_space_degree_bound_gamma
    else:
        return random_space(budget)
    result = result(
        budget=budget,
        gamma_selection_algorithm=gamma_selection_algorithm,
        point_selection_algorithm=point_selection_algorithm,
        degree_limit=degree_limit
    )
    return result


def _get_poisoning_function_feature_space(
    budget: int,
    gamma_selection_algorithm: str,
    point_selection_algorithm: str,
    degree_limit: int,
    feature_extractor,
    degree_range=None
    ):
    gamma_selection_algorithm = gammas.GAMMA_FUNCTIONS[gamma_selection_algorithm]
    if point_selection_algorithm != 'degree_range_random':
        point_selection_algorithm = selection_strategy.STRATEGIES[point_selection_algorithm]
    else:
        point_selection_algorithm = partial(
            selection_strategy.STRATEGIES[point_selection_algorithm], 
            degree_range
        )
    result = feature_space_degree_bound_gamma(
        feature_extractor=feature_extractor,
        budget=budget,
        gamma_selection_algorithm=gamma_selection_algorithm,
        point_selection_algorithm=point_selection_algorithm,
        degree_limit=degree_limit,
    )
    return result



def _get_poisoning_function(
    mode: str,
    **kwargs
    ):
    assert mode in ['projected', 'input', 'random', 'feature']
    if mode == 'random':
        return random_space(kwargs['budget'])
    elif mode in ['projected', 'input']:
        return _get_poisoning_function_input_or_projected(
                    mode=mode,
                    **kwargs
               )
    else:
        return _get_poisoning_function_feature_space(
            **kwargs
        )


def _mask_to_poisoning_function(
    mask_path: str
    ):
    assert('.pkl' in mask_path)
    with open(mask_path, 'rb') as f:
        mask = pickle.load(f)
    class Mask2PoisoningFunction(object):
        def __init__(self, mask: List[List[bool]]):
            self.counter = 0
            self.mask = mask
        def __call__(self, points):
            result = self.mask[self.counter]
            self.counter = (self.counter + 1) % len(self.mask)
            return result
    return Mask2PoisoningFunction(mask)


def poisoning_function_from_args(
    from_mask: bool,
    **kwargs
    ):
    '''
    if from_mask is True,
    then pass in 
        mask_path: str - path to the pickled mask file
    else pass in
        mode: if random, just pass in budget
        otherwise,
        budget: int,
        gamma_selection_algorithm: str,
        point_selection_algorithm: str,
        degree_limit: int,
    projected_dim: int = None 
    degree_range: (l, r)
    This will run the appropriate algorithm to produce an appropriate poisoning
    function, which can then be passed into a poisoned class in
    poisoned_dataset.py during a call to save_mask or make_loaders
    '''
    if from_mask:
        return _mask_to_poisoning_function(**kwargs)
    else:
        return _get_poisoning_function(**kwargs)

class SF:
    def __init__(self, mask, budget):
        self.counter = 0
        self.budget = budget
        self.mask = mask
    def __call__(self, points):
        mask = self.mask[self.counter]
        self.counter = (self.counter + 1) % len(self.mask)

        filtered_points = [p for p, m in zip(points, mask) if m]
        filtered_indices = [i for i, m in enumerate(mask) if m]

        # cheap fix
        if len(filtered_indices) < self.budget:
            rem = self.budget - len(filtered_indices)
            filtered_indices.extend(list(range(0, rem)))
            filtered_points.extend(points[: rem])

        # gamma, degree = gammas.search_degree_gamma(
            # filtered_points,
            # 30,
            # lower_limit=0.0, upper_limit=50.0,
            # iterations=30,
            # norm=2.0,
            # batch_size=128,
            # degree_func=gammas.GAMMA_FUNCTIONS['max'],
            # use_tqdm=True
        # )
        # print(gamma, degree)
        # adj = poisoning_adversary.get_connections(
            # points=filtered_points,
            # norm=2.0,
            # norm_threshold=gamma,
            # batch_size=128,
        # )
        # sel = selection_strategy.greedy_v2(
            # adj,
            # [self.budget],
        # )[self.budget]
        sel = poisoning_adversary.random_selector(
            filtered_points,
            [self.budget],
        )[self.budget]
        result = [False for _ in points]
        for index, s in zip(filtered_indices, sel):
            if s:
                result[index] = True
        return result


def split_by_label(x, y):
    n_labels = len(set(y))
    out = [[] for i in range(n_labels)]
    for x_, y_ in zip(x, y):
        out[y_].append(x_)
    return out


if __name__ == '__main__':
    import poisoned_datasets

    parser = ArgumentParser()
    parser.add_argument('--already_vulnerable_mask_path', type=str, required=True)
    parser.add_argument('--mask_path', type=str, required=True)
    parser.add_argument('--features', action='store_true')
    parser.add_argument('--budget', type=int, required=True)
    parser.add_argument('--seed', type=int, required=True)

    args = parser.parse_args()
    
    np.random.seed(args.seed)
    torch.manual_seed(args.seed)

    train_set = torchvision.datasets.CIFAR10(
        root='/storage/anon/datasets/cifar',
        train=True,
        transform=da.TEST_TRANSFORMS_DEFAULT(32)
    )
    x, y = poisoning_adversary.unzip(train_set)
    x = np.asarray([x_.numpy() for x_ in x])

    if args.features:
        x = pickle.load(open('/storage/anon/projected_cifar_data/features.pkl', 'rb'))

    already_vulnerable_mask = pickle.load(open(args.already_vulnerable_mask_path, 'rb'))
    
    sf = SF(budget=args.budget, mask=already_vulnerable_mask)
    result_mask = poisoning_adversary.get_poison_mask(x, y, sf)
    result_mask = split_by_label(result_mask, y)
    pickle.dump(result_mask, open(args.mask_path, 'wb'))

